Require Import List.
Require Import LibTactics.
Require Import Coq.Program.Equality.
Require Import Coq.Arith.Peano_dec.
Require Import Coq.Arith.Wf_nat.
Require Import Coq.Program.Wf.
Require Import Coq.Logic.ProofIrrelevance.
Require Import Setoid.
Require Import Metatheory.
Require Import ArithMisc.
Require Import InfiniteTypes.
Require Import Terms.
Require Import Lexicographic.
Require Import Svo.

(* TEMPORARY il serait bon d'ajouter les constantes de Fork au systeme (avec une
   distinction generique et constantes contractives et non-contractives?) et de
   montrer une propriete raffinee d'existence de formes normales de tete *)

(* ----------------------------------------------------------------------------------------------------------------- *)

(* This code was copied from Coq.Program.Wf, revision 12160 (June 2, 2009).
   It should become unnecessary with newer versions of Coq. *)

Require Import FunctionalExtensionality.

Program Lemma fix_sub_eq_ext :
    forall (A : Type) (R : A -> A -> Prop) (Rwf : well_founded R)
      (P : A -> Type)
      (F_sub : forall x : A, (forall (y : A | R y x), P y) -> P x),
      forall x : A,
        Fix_sub A R Rwf P F_sub x =
          F_sub x (fun (y : A | R y x) => Fix_sub A R Rwf P F_sub y).
  Proof.
    intros ; apply Fix_eq ; auto.
    intros.
    assert(f = g).
    extensionality y ; apply H.
    rewrite H0 ; auto.
  Qed.

Section Fix_rects.

  Set Implicit Arguments.

  Variable A: Type.
  Variable P: A -> Type.
  Variable R : A -> A -> Prop.
  Variable Rwf : well_founded R.
  Variable f: forall (x : A), (forall y: { y: A | R y x }, P (proj1_sig y)) -> P x.
  
  Lemma F_unfold x r:
    Fix_F_sub A R P f x r =
    f (fun y => Fix_F_sub A R P f (proj1_sig y) (Acc_inv r (proj2_sig y))).
  Proof. intros. case r; auto. Qed.

  (* Fix_F_sub_rect lets one prove a property of
  functions defined using Fix_F_sub by showing
  that property to be invariant over single application of the
  function body (f in our case). *)

  Lemma Fix_F_sub_rect
    (Q: forall x, P x -> Type)
    (inv: forall x: A,
     (forall (y: A) (H: R y x) (a: Acc R y),
        Q y (Fix_F_sub A R P f y a)) ->
        forall (a: Acc R x),
          Q x (f (fun y: {y: A | R y x} =>
            Fix_F_sub A R P f (proj1_sig y) (Acc_inv a (proj2_sig y)))))
    : forall x a, Q _ (Fix_F_sub A R P f x a).
  Proof with auto.
    intros Q inv.
    set (R' := fun (x: A) => forall a, Q _ (Fix_F_sub A R P f x a)).
    cut (forall x, R' x)...
    apply (well_founded_induction_type Rwf).
    subst R'.
    simpl.
    intros.
    rewrite F_unfold...
  Qed.

  (* Finally, Fix_F_rect lets one prove a property of
  functions defined using Fix_F by showing that
  property to be invariant over single application of the function
  body (f). *)

  Lemma Fix_sub_rect
    (Q: forall x, P x -> Type)
    (inv: forall
      (x: A)
      (H: forall (y: A), R y x -> Q y (Fix_sub A R Rwf P f y))
      (a: Acc R x),
        Q x (f (fun y: {y: A | R y x} => Fix_sub A R Rwf P f (proj1_sig y))))
    : forall x, Q _ (Fix_sub A R Rwf P f x).
  Proof with auto.
    unfold Fix_sub.
    intros.
    apply Fix_F_sub_rect.
    intros.
    assert (forall y: A, R y x0 -> Q y (Fix_F_sub A R P f y (Rwf y)))...
    set (inv x0 X0 a). clearbody q.
    (* fpottier *)
    replace
      (fun y: {y: A | R y x0} => Fix_F_sub A R P f (proj1_sig y) (Acc_inv a (proj2_sig y)))
    with
      (fun y: {y: A | R y x0} => Fix_F_sub A R P f (proj1_sig y) (Rwf (proj1_sig y))).
    assumption.
    extensionality y. myEqualUpToPI.
  Qed.

End Fix_rects.

(* ----------------------------------------------------------------------------------------------------------------- *)

(* The interpretation of types. *)

(* We use the word ``candidate'' for a set of lambda-terms. *)

Definition candidate :=
  trm -> Prop.

Definition candidate_inclusion : candidate -> candidate -> Prop :=
  fun c1 c2 => forall T, c1 T -> c2 T.

Hint Unfold candidate_inclusion.

(* The full candidate is the set of all (well-formed) lambda-terms. *)

Definition full : candidate :=
  term.

(* The interpretation of the type [TyLeaf] is the set of all nice terms. *)

(* This fact is not used often -- for many purposes, [cleaf] can be viewed
   as opaque -- so we explicitly unfold this definition where needed. *)

Definition cleaf : candidate :=
  nice.

Lemma cleaf_term:
  candidate_inclusion cleaf term.
Proof.
  eauto using nice_term.
Qed.

(* The interpretation of types cannot be defined by structural induction over
   types, because types are infinite. Neither can it be directly defined by
   induction over integer indices, because the integer index does not decrease
   at every recursive call -- it only decreases at bullets. To address these
   problems, we use a definition in two steps. *)

(* This is the first step in the definition of the interpretation of types.
   We assume that the type [t] is well-formed, which implies that it is also
   bullet-bounded: every path through [t] must reach a leaf or a bullet.  We
   define the interpretation of [t] by structural induction over a proof of
   [bb t]. This means that we must stop at bullets: there, we invoke another
   interpretation, represented by the parameter [later]. *)

Lemma wf_eta:
  forall A, forall R : A -> A -> Prop, well_founded R -> well_founded (fun x => R x).
Proof.
  intros.
  intro.
  assert (forall a, Acc R a -> Acc (fun x => R x) a).
    induction 1. constructor. intros. auto.
  eauto.
Qed.

Definition pair : forall t, wf t -> { t : ty | wf t } :=
  exist (fun t => wf t).

Inductive subterm : { t : ty | wf t } -> { t : ty | wf t } -> Prop :=
| SubtermDomain:   forall t u v wft wfu, t = TyArrow u v -> subterm (pair u wfu) (pair t wft)
| SubtermCodomain: forall t u v wft wfv, t = TyArrow u v -> subterm (pair v wfv) (pair t wft).

Lemma subterm_accessibility:
  forall t   : ty,
  forall wft : wf t,
  forall bbt : bb t,
  Acc subterm (pair t wft).
Proof.
  intros ? ? bbt. induction bbt; constructor; intros ? h; dependent destruction h; eauto.
Qed.

Lemma subterm_wf: well_founded subterm.
Proof.
  intros [ ? ? ]. apply subterm_accessibility. apply wf_bb. auto.
Qed.

Definition R : nat * { t : ty | wf t } -> nat * { t : ty | wf t } -> Prop :=
  lexprod _ _ lt subterm.

Lemma wfR: well_founded R.
Proof.
  apply lexprod_wf.
  apply lt_wf.
  apply subterm_wf.
Qed.

Program Fixpoint interpret (it : nat * { t : ty | wf t }) { wf R it } : candidate :=
  match it with
  | (S j, t) =>
      match t with
      | TyLeaf =>
          cleaf
      | TyArrow u v =>
          fun V : trm =>
            term V /\
            forall k, k <= S j ->
            forall U,
            (interpret (k, u) U) ->
            (interpret (k, v) (trm_app V U))
      | TyBullet u =>
          interpret (j, u)
      end
  | (0, _) =>
      full
  end.

Next Obligation.
  eauto using wf_hereditary_domain.
Defined.

Next Obligation.
  rewrite <- Heq_it.
  case (eq_nat_dec k (S j)); intro.
    (* k = S j *)
    subst k. eapply LexprodRight. eapply SubtermDomain. symmetry. eassumption.
    (* k < S j *)
    eapply LexprodLeft. omega.
Qed.

Next Obligation.
  eauto using wf_hereditary_codomain.
Defined.

Next Obligation.
  rewrite <- Heq_it.
  case (eq_nat_dec k (S j)); intro.
    (* k = S j *)
    subst k. eapply LexprodRight. eapply SubtermCodomain. symmetry. eassumption.
    (* k < S j *)
    eapply LexprodLeft. omega.
Qed.

Next Obligation.
  rewrite <- Heq_it.
  apply LexprodLeft.
  auto.
Qed.

Next Obligation.
  apply wf_eta.
  apply wfR.
Defined.

(* ---------------------------------------------------------------------------------------------------------------------- *)

(* We now prove that [interpret] satisfies the expected equations. *)

Lemma interpret_pi:
  forall i t wft1 wft2 T,
  interpret (i, pair t wft1) T ->
  interpret (i, pair t wft2) T.
Proof.
  intros. myPunchPI wft2. eassumption.
Qed.

Hint Immediate interpret_pi.

Lemma interpret_zero:
  forall t T,
  term T ->
  interpret (0, t) T.
Proof.
  intros.
  unfold interpret; rewrite fix_sub_eq_ext; simpl. auto.
Qed.

Lemma interpret_zero_back:
  forall t T,
  interpret (0, t) T ->
  term T.
Proof.
  intros ? ?.
  unfold interpret; rewrite fix_sub_eq_ext; simpl. auto.
Qed.

Lemma interpret_leaf:
  forall i wft,
  interpret (S i, pair TyLeaf wft) = cleaf.
Proof.
  intros. unfold interpret. rewrite fix_sub_eq_ext; simpl. auto.
Qed.

Lemma interpret_arrow:
  forall j u v wft V,
  let wfu := wf_hereditary_domain _ wft _ _ refl in
  let wfv := wf_hereditary_codomain _ wft _ _ refl in
  interpret (S j, pair (TyArrow u v) wft) V ->
  term V /\
  forall k, k <= S j ->
  forall U,
  (interpret (k, pair u wfu) U) ->
  (interpret (k, pair v wfv) (trm_app V U)).
Proof.
  intros ? ? ? ? ? ? ?.
  myPunchPI wfu.
  myPunchPI wfv.
  unfold interpret. rewrite fix_sub_eq_ext. simpl.
  intros. eauto.
Qed.

Lemma interpret_arrow_back:
  forall j u v wft wfu wfv V,
  term V ->
  (forall k, k <= S j -> forall U, interpret (k, pair u wfu) U -> interpret (k, pair v wfv) (trm_app V U)) ->
  interpret (S j, pair (TyArrow u v) wft) V.
Proof.
  intros ? ? ? ? ? ? ?.
  myPunchPI wfu.
  myPunchPI wfv.
  intros ? H.
  unfold interpret. rewrite fix_sub_eq_ext. simpl.
  split; auto.
  intros k ksj U.
  generalize (H k ksj U); clear H; intro H.
  unfold interpret in H. apply H.
Qed.

Lemma interpret_bullet:
  forall j u wft,
  interpret (S j, pair (TyBullet u) wft) = interpret (j, pair u (everywhere_hereditary_bullet bb u wft)).
Proof.
  intros. unfold interpret. rewrite fix_sub_eq_ext. simpl. reflexivity. (* miracle! *)
Qed.

Lemma interpret_zero_zero:
  forall t u T,
  interpret (0, t) T ->
  interpret (0, u) T.
Proof.
  intros. apply interpret_zero. eapply interpret_zero_back. eassumption.
Qed.

Lemma interpret_bullet_general:
  forall u wfu j T,
  interpret (j, pair (TyBullet u) (wf_bullet _ wfu)) T ->
  interpret (j - 1, pair u wfu) T.
Proof.
  introv h; dependent destruction j.
  eauto using interpret_zero_zero.
  rewrite arith_successor_minus_one. rewrite interpret_bullet in h. myPunchPI wfu. eassumption.
Qed.

Lemma interpret_bullet_general_back:
  forall j u wft T,
  interpret (j - 1, pair u (everywhere_hereditary_bullet bb u wft)) T ->
  interpret (j, pair (TyBullet u) wft) T.
Proof.
  introv h; dependent destruction j.
  eauto using interpret_zero_zero.
  rewrite arith_successor_minus_one in h. rewrite interpret_bullet. assumption.
Qed.

Lemma pair_equality:
  forall t u wft wfu,
  t = u ->
  pair t wft = pair u wfu.
Proof.
  intros. apply sig_equality. auto.
Qed.

Lemma interpret_bullets:
  forall k j u wfu T,
  interpret (j, pair (bullets k u) (wf_bullets _ _ wfu)) T ->
  interpret (j - k, pair u wfu) T.
Proof.
  induction k; simpl; introv h.
  rewrite arith_minus_zero. eauto.
  rewrite arith_minus_successor.
  rewrite (pair_equality _ _ _ _ (reverse_bullets k u)) in h.
  forwards: interpret_bullet_general. eexact h.
  apply IHk. eassumption.
Qed.

Lemma interpret_bullets_back:
  forall k j u wft T,
  interpret (j - k, pair u (wf_hereditary_bullets _ _ wft)) T ->
  interpret (j, pair (bullets k u) wft) T.
Proof.
  induction k; simpl; introv h.
  rewrite arith_minus_zero in h. eauto. 
  rewrite arith_minus_successor in h.
  assert (wf (TyBullet (bullets k u))) as wfbku.
    assert (wf u).
      eapply wf_hereditary_bullet.
      eapply wf_hereditary_bullets.
      eauto. eauto.
    eauto.
  rewrite (pair_equality _ _ _ wfbku (reverse_bullets k u)).
  apply interpret_bullet_general_back. eapply IHk. eauto.
Qed.

(* ---------------------------------------------------------------------------------------------------------------------- *)

(* Basic properties of the interpretation of types. *)

Lemma interpret_term:
  forall it T,
  interpret it T ->
  term T.
Proof.
  intro it; destruct it as [ i tt ]. generalize tt. induction i.
  (* i = 0 *)
  apply interpret_zero_back.
  (* i > 0 *)
  intro t. dependent destruction t. dependent destruction x; simpl; intros.
  (* Leaf *)
  rewrite interpret_leaf in H. apply cleaf_term. auto.
  (* Arrow *)
  eapply interpret_arrow. eassumption.
  (* Bullet *)
  rewrite interpret_bullet in H.
  eauto.
Qed.

Hint Extern 1 (term ?T) =>
  match goal with
  | H: interpret ?it T |- _ => exact (interpret_term it T H)
  end.

Lemma interpret_arrow_general:
  forall j u v wft V,
  interpret (j, pair (TyArrow u v) wft) V ->
  forall k, k <= j ->
  forall U,
  (interpret (k, pair u (wf_hereditary_domain _ wft _ _ refl)) U) ->
  (interpret (k, pair v (wf_hereditary_codomain _ wft _ _ refl)) (trm_app V U)).
Proof.
  intro j; dependent destruction j.
  (* j = 0 *)
  intros. assert (k = 0). omega. subst.
  apply interpret_zero.
  constructor; eauto.
  (* j > 0 *)
  intros. eapply interpret_arrow; eassumption.
Qed.

Lemma interpret_arrow_back_general:
  forall v wfv j u wft wfu V,
  term V ->
  (forall k, k <= j -> forall U, interpret (k, pair u wfu) U -> interpret (k, pair v wfv) (trm_app V U)) ->
  interpret (j, pair (TyArrow u v) wft) V.
Proof.
  intros ? ? j; dependent destruction j.
  (* j = 0 *)
  intros. eauto using interpret_zero.
  (* j > 0 *)
  intros. eapply interpret_arrow_back; eauto.
Qed.

Lemma interpret_zero_general:
  forall i t u T,
  interpret (i, t) T ->
  interpret (0, u) T.
Proof.
  intros. apply interpret_zero. eauto.
Qed.

Lemma interpret_monotonic:
  forall i j, j <= i ->
  forall t T,
  interpret (i, t) T ->
  interpret (j, t) T.
Proof.
  induction i.
  (* i = 0 *)
  intros. assert (j = 0). omega. subst j. assumption.
  (* i > 0 *)
  intros.
  case (eq_nat_dec j (S i)).
    (* j = S i *)
    intro. subst j. assumption.
    (* j <= i *)
    intro. assert (j <= i). omega. apply IHi. auto.
    clear j H n H1.
    (* Here comes the bulk of the proof: we have [interpret (S i, t) T]
       and must show [interpret (i, t) T]. *)
    dependent destruction i.
      (* i = 0 *)
      apply interpret_zero. eauto.
      (* i > 0 *)
      dependent destruction t. dependent destruction x.
      (* Leaf *)
      rewrite interpret_leaf in H0. rewrite interpret_leaf. auto.
      (* Arrow *)
      eapply interpret_arrow_back.
      eapply interpret_arrow. eassumption.
      intros k ? U ?.
      eapply interpret_arrow. eassumption. omega. eassumption.
      (* Bullet *)
      rewrite interpret_bullet in H0. rewrite interpret_bullet.
      eauto.
Qed.

Lemma interpret_monotonic_bis:
  forall i j t T,
  interpret (i, t) T ->
  j <= i ->
  interpret (j, t) T.
Proof.
  intros. eapply interpret_monotonic; eassumption.
Qed.

Hint Extern 2 (interpret (?i, ?t) ?T) =>
  eapply interpret_monotonic_bis; [ (eassumption || (eapply interpret_pi; eassumption)) | omega ].

Lemma interpret_bullet_intro:
  forall i t T wft wfbt,
  interpret (i, pair t wft) T ->
  interpret (i, pair (TyBullet t) wfbt) T.
Proof.
  introv h. apply interpret_bullet_general_back. eauto.
Qed.

Lemma interpret_bullets_intro:
  forall k i t T wft wfbst,
  interpret (i, pair t wft) T ->
  interpret (i, pair (bullets k t) wfbst) T.
Proof.
  induction k; intros; simpl.
  myPunchPI wfbst. eassumption.
  assert (wf (TyBullet t)) as wfbt. eauto.
  eapply (IHk _ _ _ wfbt). eapply interpret_bullet_intro; eauto.
Qed.

(* ---------------------------------------------------------------------------------------------------------------------- *)

(* The model validates subtyping. *)

Definition subg q i t j u :=
  forall p wft wfu T, p <= q -> interpret (p - i, pair t wft) T -> interpret (p - j, pair u wfu) T.

Lemma subg_zero:
  forall q i t j u,
  q <= j ->
  subg q i t j u.
Proof.
  introv ? ? h.
  replace (p - j) with 0. apply interpret_zero. eauto. omega.
Qed.

Lemma subg_successor:
  forall q i t j u,
  subg q i t j u ->
  subg (q + 1) (i + 1) t (j + 1) u.
Proof.
  introv ss ? h.
  unfold subg in ss.
  rewrite arith_minus_plus.
  eapply ss. omega.
  rewrite <- arith_minus_plus.
  auto.
Qed.

Ltac myTrivialIfZero i :=
  let h := fresh in
  case (eq_nat_dec i 0); intro h; [ rewrite h; apply interpret_zero; auto | ].

Lemma subg_successor_p:
  forall q i t j u,
  subg q (i - 1) t (j - 1) u ->
  j > 0 ->
  subg (S q) i t j u.
Proof.
  introv ss ? ? h.
  unfold subg in ss.
  eapply interpret_monotonic_bis.
  eapply (ss (p - 1)). omega. eauto.
  omega.
Qed.

Lemma subg_leaf:
  forall q i j,
    i <= j ->
    subg q i TyLeaf j TyLeaf.
Proof.
  unfold subg. intros. eauto.
Qed.

Lemma subg_arrow:
  forall q k i t1 t2 j u1 u2,
    subg q j u1 (i + k) t1 ->
    subg q (i + k) t2 j u2 ->
    subg q i (TyArrow t1 t2) j (TyArrow u1 u2).
Proof.
  introv domain codomain ? h.
  assert (wf u1) as wfu1. eauto.
  assert (wf u2) as wfu2. eauto.
  eapply (interpret_arrow_back_general _ wfu2). eauto. intros o ? ? u1U.
  (* If [o] is zero, the result is immediate. *)
  myTrivialIfZero o.
  (* We now assume [o > 0]. This ensures that the subtractions do not hit zero. *)
  unfold subg in domain.
  forwards: (domain (o + j) wfu1). omega.
  eauto.
  forwards: interpret_arrow_general. eexact h. instantiate (1 := (o + j) - (i + k)). omega.
  eauto.
  unfold subg in codomain.
  forwards: (codomain (o + j)). omega.
  eauto.
  eauto.
Qed.

Lemma subg_bullet_left:
  forall q i t j u,
  subg q (i + 1) t j u ->
  subg q i (TyBullet t) j u.
Proof.
  unfold subg. introv s ? h.
  assert (wf t) as wfh. eauto.
  forwards: (interpret_bullet_general _ wfh). eauto. eauto.
Qed.

Lemma subg_bullets_left:
  forall k q i t j u,
  subg q (i + k) t j u ->
  subg q i (bullets k t) j u.
Proof.
  induction k; simpl; introv h.
  rewrite arith_plus_zero in h. auto.
  rewrite reverse_bullets. apply subg_bullet_left. apply IHk. replace (i + 1 + k) with (i + S k); try omega. auto.
Qed.

Lemma subg_bullet_right:
  forall q i t j u,
  subg q i t (j + 1) u ->
  subg q i t j (TyBullet u).
Proof.
  unfold subg. introv s ? h.
  forwards: interpret_bullet_general_back. 
  eapply interpret_monotonic_bis. eapply s. eassumption. eassumption.
  instantiate (1 := p - j). omega. eassumption.
Qed.

Lemma subg_bullets_right:
  forall k q i t j u,
  subg q i t (j + k) u ->
  subg q i t j (bullets k u).
Proof.
  induction k; simpl; introv h.
  rewrite arith_plus_zero in h. auto.
  rewrite reverse_bullets. apply subg_bullet_right. apply IHk. replace (j + 1 + k) with (j + S k); try omega. auto.
Qed.

Lemma subtyping_valid_preliminary:
  forall q (itju : quadruple),
  let (itj, u) := itju in
  let (it, j) := itj in
  let (i, t) := it in
  forall ft fu,
  wf t -> wf u ->
  sksub i (sk 0 t ft) j (sk 0 u fu) ->
  subg q i t j u.
Proof.
  intro q.
  induction itju as [ itju IH ] using (well_founded_ind (svo_wf q)).
  destruct itju as [[[ i t ] j ] u ].
  introv ? ? ss.
  (* If [q <= j] holds, then the result is trivial, and does not
     require the induction hypothesis. *)
  case (eq_nat_dec (q - j) 0); intro.
  apply subg_zero. omega.
  (* We may now assume [j < q]. *)
  dependent destruction u.
  (* Leaf *)
  dependent destruction t.
    (* Leaf Leaf *)
    rewrite sk_leaf in ss.
    rewrite sk_leaf in ss.
    dependent destruction ss.
    apply subg_leaf. omega.
    (* Arrow Leaf (impossible) *)
    rewrite sk_leaf in ss.
    rewrite sk_arrow in ss.
    dependent destruction ss.
    (* Bullet Leaf *)
    rewrite sk_bullet in ss.
    forwards ssp: sksub_plus_sk. eexact ss. rewrite arith_plus_zero in ssp.
    apply subg_bullet_left.
    eapply (IH (i + 1, t, j, TyLeaf)); eauto. apply SvoBL; constructor; eauto. omega.
  (* Arrow *)
  dependent destruction t.
    (* Leaf Arrow (impossible) *)
    rewrite sk_leaf in ss.
    rewrite sk_arrow in ss.
    dependent destruction ss.
    (* Arrow Arrow *)
    rewrite sk_arrow in ss.
    rewrite sk_arrow in ss.
    dependent destruction ss.
    eapply subg_arrow.
    eapply (IH (j, u1, i + k, t1)); eauto. apply SvoAL; constructor; eauto. omega.
    eapply (IH (i + k, t2, j, u2)); eauto. apply SvoAR; constructor; eauto. omega.
    (* Bullet Arrow (identical to Bullet Leaf above) *)
    rewrite sk_bullet in ss.
    forwards ssp: sksub_plus_sk. eexact ss. rewrite arith_plus_zero in ssp.
    apply subg_bullet_left.
    eapply (IH (i + 1, t, j, TyArrow u1 u2)); eauto. apply SvoBL; constructor; eauto. omega.
  (* Bullet *)
  apply subg_bullet_right.
  rewrite sk_bullet in ss.
  forwards ssp: sksub_plus_sk. eexact ss. rewrite arith_plus_zero in ssp.
  eapply (IH (i, t, j + 1, u)); eauto. apply SvoBR; constructor; eauto. omega.
Qed.

Lemma subtyping_valid_general:
  forall q i t j u ft fu,
  wf t -> wf u ->
  sksub i (sk 0 t ft) j (sk 0 u fu) ->
  subg q i t j u.
Proof.
  intros. eapply (subtyping_valid_preliminary q (i, t, j, u)); eauto.
Qed.

Lemma subtyping_valid:
  forall t u wft wfu,
  sub t u ->
  forall i T,
  interpret (i, pair t wft) T ->
  interpret (i, pair u wfu) T.
Proof.
  introv ss h.
  dependent destruction ss.
  forwards soub: (subtyping_valid_general i). eexact wft. eexact wfu. eassumption.
  unfold subg in soub.
  forwards uT: (soub i). omega. rewrite arith_minus_zero. eassumption.
  rewrite arith_minus_zero in uT.
  eassumption.
Qed.

(* ----------------------------------------------------------------------------------------------------------------- *)

(* A definition of what it means for a set [S] to be closed with
  respect to a relation [R]. *)

Definition closed (step : trm -> trm -> Prop) (S : trm -> Prop) :=
  forall T,
  S T ->
  forall U,
  step T U ->
  S U.

(* The interpretation of types is closed with respect to a relation [step], as
   soon as: (i) the interpretation of the basic type [TyLeaf] is closed with
   respect to [step], and (ii) [step] is a congruence with respect to the left
   application context. *)

Lemma interpret_closed:
  forall step : trm -> trm -> Prop,
  closed step cleaf ->
  (forall T U, step T U -> forall V, term V -> step (trm_app T V) (trm_app U V)) ->
  (forall T U, step T U -> term U) ->
  forall it,
  closed step (interpret it).
Proof.
  intros ? Hleaf Hcongruence Hregular.
  induction it as [[ i [ t wft ]] IH ] using (well_founded_ind wfR).
  intros T itT U stepTU.
  dependent destruction i.
  (* i = 0 *)
  apply interpret_zero. eauto.
  (* i > 0 *)
  dependent destruction t.
  (* Leaf *)
  rewrite interpret_leaf.
  rewrite interpret_leaf in itT.
  eauto.
  (* Arrow *)
  assert (wft1: wf t1).
    eauto using wf_hereditary_domain.
  assert (wft2: wf t2).
    eauto using wf_hereditary_codomain.
  apply interpret_arrow_back with (wfu := wft1) (wfv := wft2). eauto.
  intros ? ? V it1V.
  assert (IHt2: closed step (interpret (k, pair t2 wft2))).
    apply IH.
    unfold R.
    apply lexprod_intro.
    omega.
    eapply SubtermCodomain.
    reflexivity.
  unfold closed in IHt2.
  apply IHt2 with (T := (trm_app T V)).
  myPunchPI wft1 in it1V.
  myPunchPI wft2.
  eapply interpret_arrow; eassumption.
  eauto.
  (* Bullet *)
  rewrite interpret_bullet.
  rewrite interpret_bullet in itT.
  apply IH with (T := T).
  apply LexprodLeft. omega.
  auto. auto.
Qed.

(* In particular, provided the interpretation of [TyLeaf] is closed with
   respect to reduction, then so is the interpretation of every type. *)

Lemma interpret_red:
  closed red cleaf ->
  forall it,
  closed red (interpret it).
Proof.
  eauto using interpret_closed.
Qed.

(* The interpretation of types is closed with respect to the inverse of a
   relation [step], as soon as: (i) the interpretation of the basic type
   [TyLeaf] is closed with respect to the inverse of [step], and (ii) [step]
   is a congruence with respect to the left application context. *)

Lemma interpret_closed_flip:
  forall step : trm -> trm -> Prop,
  closed (flip step) cleaf ->
  (forall T U, step T U -> forall V, term V -> step (trm_app T V) (trm_app U V)) ->
  (forall T U, step T U -> term T) ->
  forall it,
  closed (flip step) (interpret it).
Proof.
  intros ? Hleaf Hcongruence Hregular.
  induction it as [[ i [ t wft ]] IH ] using (well_founded_ind wfR).
  intros T itT U stepUT.
  unfold flip in stepUT.
  dependent destruction i.
  (* i = 0 *)
  apply interpret_zero. eauto.
  (* i > 0 *)
  dependent destruction t.
  (* Leaf *)
  rewrite interpret_leaf.
  rewrite interpret_leaf in itT.
  eauto.
  (* Arrow *)
  assert (wft1: wf t1).
    eauto using wf_hereditary_domain.
  assert (wft2: wf t2).
    eauto using wf_hereditary_codomain.
  apply interpret_arrow_back with (wfu := wft1) (wfv := wft2). eauto.
  intros ? ? V it1V.
  assert (IHt2: closed (flip step) (interpret (k, pair t2 wft2))).
    apply IH.
    unfold R.
    apply lexprod_intro.
    omega.
    eapply SubtermCodomain.
    reflexivity.
  unfold closed, flip in IHt2.
  apply IHt2 with (T := (trm_app T V)).
  myPunchPI wft1 in it1V.
  myPunchPI wft2.
  eapply interpret_arrow; eassumption.
  eauto.
  (* Bullet *)
  rewrite interpret_bullet.
  rewrite interpret_bullet in itT.
  apply IH with (T := T).
  apply LexprodLeft. omega.
  auto. auto.
Qed.

(* In particular, provided the interpretation of [TyLeaf] is closed with
   respect to the inverse of leftmost outermost reduction, then so is the
   interpretation of every type. *)

Lemma interpret_lored:
  closed (flip lored) cleaf ->
  forall it,
  closed (flip lored) (interpret it).
Proof.
  eauto using interpret_closed_flip.
Qed.

(* The interpretation of [TyLeaf] is the set of all nice terms, which
   is fact closed with respect to the inverse of leftmost outermost
   reduction, so we have the following corollary. *)

Lemma interpret_lored_corollary:
  forall it T U,
  lored T U ->
  interpret it U ->
  interpret it T.
Proof.
  intros.
  eapply interpret_lored; eauto.
  unfold closed, flip, cleaf. eauto using nice_closed_flip_lored.
Qed.

(* ----------------------------------------------------------------------------------------------------------------- *)

(* If the interpretation of [TyLeaf] is stable under applications to
   arbitrary terms, then it is a subset of the interpretation of any
   type. *)

Lemma cleaf_core:
  (forall U, cleaf U -> forall V, term V -> cleaf (trm_app U V)) ->
  forall it T,
  cleaf T ->
  interpret it T.
Proof.
  intro Hstable.
  induction it as [[ i [ t wft ]] IH ] using (well_founded_ind wfR).
  dependent destruction i.
  (* i = 0 *)
  intros. apply interpret_zero. eauto using cleaf_term.
  (* i > 0 *)
  dependent destruction t.
  (* Leaf *)
  rewrite interpret_leaf. auto.
  (* Arrow *)
  intros.
  assert (wft1: wf t1).
    eauto using wf_hereditary_domain.
  assert (wft2: wf t2).
    eauto using wf_hereditary_codomain.
  apply interpret_arrow_back with (wfu := wft1) (wfv := wft2). eauto using cleaf_term.
  intros ? ? V it1V.
  apply IH.
  apply lexprod_intro. omega. eapply SubtermCodomain. reflexivity.
  apply Hstable; eauto.
  (* Bullet *)
  rewrite interpret_bullet.
  apply IH.
  apply LexprodLeft.
  omega.
Qed.

(* The interpretation of [TyLeaf] is the set of all nice terms, which
   is in fact stable under applications to arbitrary terms, so we have
   the following corollary. *)

Lemma nice_core:
  forall it T,
  nice T ->
  interpret it T.
Proof.
  apply cleaf_core.
  unfold cleaf. eauto using nice_stable.
Qed.

(* ----------------------------------------------------------------------------------------------------------------- *)

(* Let us try and sum up the requirements that bear on [cleaf], the
   interpretation of the type [TyLeaf].

   It must be contain the variables, and be stable under applications to
   arbitrary terms, so that (thanks to [cleaf_core] above) we can claim that
   the interpretation of any type includes the variables.  (This fact is later
   used to summon a fresh variable at an arbitrary type.) As a result, [cleaf]
   must include all terms of the form [x T ... T].

   It must be closed under inverse reduction, for a notion of reduction that
   includes beta and is a congruence with respect to the the left of an
   application. (This fact is later used to prove that, if [T/x]U inhabits the
   interpretation of a type, then so does (\x.U)T.) In other words, it must be
   closed with respect to the inverse of leftmost outermost reduction.

   It seems natural that it be closed under arbitrary reductions, because
   reduction can only improve the behavior of a term. I don't know yet whether
   this property is required. Maybe it is just a consequence of the above
   definitions: take [cleaf] to be the set of terms that leftmost outermost
   reduce to a term of the form [x T ... T]; this set seems closed under
   arbitrary reductions.

   It should contain only terms that reduce to a weak head normal form,
   because the purpose of the entire construction is to prove that well-typed
   terms have a weak head normal form.

   As a consequence of the above requirements, it seems that [cleaf] must not
   contain the lambda-abstractions. Indeed, if it contains a
   lambda-abstraction \x.U, then, because it is closed under applications, it
   must contain (\x.U)T as well. Because it is closed under reduction, it must
   then contain [T/x]U -- an arbitrary term, which perhaps does not have a
   weak head normal form. *)

(* ----------------------------------------------------------------------------------------------------------------- *)

(* The model validates the typing judgement. *)

(* A substitution [sigma] obeys a typing environment [E] at level [i]
   iff, for every variable [x], the term associated with [x] through
   [sigma] lies in in the interpretation of the type associated with
   [x] through [E]. *)

Definition obeys (i : nat) (sigma : substitution) (E : env ty) : Prop :=
  forall x t wft,
  binds x t E ->
  interpret (i, pair t wft) (isubst sigma (trm_fvar x)).

(* [obeys] is monotonic in [i]. *)

Lemma obeys_monotonic:
  forall i j sigma E,
  obeys i sigma E ->
  j <= i ->
  obeys j sigma E.
Proof.
  introv ob ji. intros ? ? ? ?.
  apply interpret_monotonic_bis with (i := i); auto.
Qed.

(* Under suitable freshness conditions, [obeys] is preserved when
   extending both [sigma] and [E] with consistent new bindings. *)

Lemma obeys_cons:
  forall i sigma E,
  obeys i sigma E ->
  forall x t wft V,
  x \notin (subst_domain sigma) ->
  x \notin (subst_codomain sigma) ->
  interpret (i, pair t wft) V ->
  obeys i ((x, V) :: sigma) (E & x ~ t).
Proof.
  introv ob xdo xcodo itV. intros z u wfu zu.
  binds_case zu as foo bar.
  (* z is bound here, i.e. z is x. *)
  generalize (binds_single_inv bar); intros [ ? ? ]; subst. simpl. rewrite isubst_fresh; auto. rewrite subst_self. auto.
  (* z is bound further. *)
  simpl.
  (* claim that the substitution [x ~> V] is useless *)
  rewrite subst_fresh.
  (* exploit the claim *)
  apply ob. eauto.
  (* now prove the claim *)
  apply fv_isubst. eauto. simpl. rewrite dom_single in Fr. auto.
Qed.

(* The main theorem. *)

Lemma typing_valid:
  forall E T t wft,
  jty E T t ->
  term T -> (* redundant with the above assumption, but helps *)
  forall i,
  forall sigma,
  subst_valid sigma ->
  obeys i sigma E ->
  interpret (i, pair t wft) (isubst sigma T).
Proof.
  induction 1 as [ | ? ? ? ? ? ? ? ? IH | |]; introv tT ? ob.

  (* Variable. *)
  apply ob. auto.

  (* Abstraction. *)
  rewrite isubst_abs.
  assert (wfdom: wf t).
    eauto using wf_hereditary_domain.
  assert (wfcodom: wf u).
    eauto using wf_hereditary_codomain.
  eapply interpret_arrow_back_general with (wfu := wfdom) (wfv := wfcodom); eauto.
  intros ? ? V ?.
  assert (woo: term (trm_abs (isubst sigma e))).
    eauto.
  assert (her: term V).
    eauto.
  (* We have a beta-redex. Exhibit the reduction step. *)
  generalize (lored_beta woo her); intro step.
  (* Pick a fresh name. *)
  pick_fresh x.
  rewrite (@isubst_intro x) in step; eauto.
  (* Change the goal to proving that the reduct is in the interpretation. *)
  eapply interpret_lored_corollary. eassumption.
  (* Apply the induction hypothesis. *)
  apply IH; eauto.
  (* Discharge the remaining side conditions. *)
  eapply obeys_cons; eauto. eauto using obeys_monotonic.
  eauto using fv_isubst.
  (* Yay! *)

  (* Application. *)
  rewrite isubst_app.
  assert (wftu: wf (TyArrow t u)).
    eauto.
  myPunchPI wft.
  eapply (interpret_arrow_general i) with (wft := wftu).
  eauto. omega. eauto.

  (* Subtyping. *)
  assert (woo: wf t).
    eauto.
  apply subtyping_valid with (wft := woo); eauto.

Qed.

(* ----------------------------------------------------------------------------------------------------------------- *)

(* The identity substitution obeys the empty environment. *)

Lemma obeys_nil_empty:
  forall i,
  obeys i nil empty.
Proof.
  intros ? ? ? ? ?.
  false. eauto using binds_empty_inv.
Qed.

(* In fact, the identity substitution obeys every environment. This holds
   thanks to the fact that the variables inhabit the interpretation of every
   type. *)

Lemma obeys_nil:
  forall i E,
  obeys i nil E.
Proof.
  intros ? ? ? ? ? _.
  simpl.
  apply nice_core.
  eauto using nice_inactive.
Qed.

(* We are now able to apply the main theorem to the identity substitution.
   This yields a simple statement of the fact that the model validates the
   type system. *)

Theorem typing_valid_corollary:
  forall E T t wft,
  jty E T t ->
  forall i,
  interpret (i, pair t wft) T.
Proof.
  intros.
  replace T with (isubst nil T); auto.
  eapply typing_valid with (sigma := nil); eauto using obeys_nil.
Qed.

(* We now specialize this statement to the type [TyLeaf]. This allows us
   to prove that every term of type [TyLeaf] is nice, that is, every such
   term reduces to an inactive term. *)

Lemma leaf_nice:
  forall E T,
  jty E T TyLeaf ->
  nice T.
Proof.
  intros ? ? j.
  generalize (typing_valid_corollary E T TyLeaf wf_leaf j 1).
  rewrite interpret_leaf.
  auto.
Qed.

(* ----------------------------------------------------------------------------------------------------------------- *)

(* None of the above proofs are explicitly present in Nakano's paper.  We are
   now back to a point where Nakano does give a proof. (See, for instance, the
   appendix of his 2001 paper.) He proves that, for any tail-finite type [t],
   every term in the interpretation of [t] admits a head normal form. The proof
   is by induction on the tail-finiteness judgement. *)

Theorem interpret_hnf:
  forall t (wft : wf t),
  tf t ->
  forall T,
  (forall i wft, interpret (i, pair t wft) T) ->
  reducible T.
Proof.
  induction 2; introv itT.

  (* Leaf *)
  apply nice_reducible.
  generalize (itT 1 wf_leaf). rewrite interpret_leaf.
  auto.

  (* Arrow *)
  (* Pick a fresh variable [y]. *)
  pick_fresh y.
  (* The term [T y] is in the interpretation of the codomain [v]. *)
  assert (forall i wfv, interpret (i, pair v wfv) (trm_app T (trm_fvar y))).
    intros.
    assert (wfuv: wf (TyArrow u v)).
      eauto.
    myPunchPI wfv.
    apply interpret_arrow_general with (j := i) (wft := wfuv); auto.
    eauto using nice_core.
  (* Therefore, by the induction hypothesis, [T y] admits a head normal form. *)
  assert (reducible (trm_app T (trm_fvar y))).
    eauto.
  (* There follows that [T] itself admits a head normal form. *)
  eauto using reducible_app_inv.

  (* Bullet *)
  (* Apply the induction hypothesis *)
  apply IHtf; eauto. intros.
  (* Exploit the fact that the (toplevel) interpretations of [t] and [TyBullet t]
     are the same. *)
  assert (wfbt: wf (TyBullet t)). eauto.
  generalize (itT (S i) wfbt). rewrite interpret_bullet.
  eauto.

Qed.

Definition typed (T : trm) :=
  exists E, exists t,
  jty E T t.

Theorem typed_hnf:
  forall T,
  typed T ->
  reducible T.
Proof.
  intros ? [ E [ t j ]].
  apply interpret_hnf with (t := t); eauto.
  eauto using typing_valid_corollary.
Qed.

(* ----------------------------------------------------------------------------------------------------------------- *)

(* We now define what it means for a term to admit a total (or maximal) Boehm
   tree. *)

(* Ideally, this should be a mixture of co-inductive and inductive
   definitions, because reducing to a head normal form is an inductive notion,
   whereas the totality of a Boehm tree is a co-inductive notion. However,
   Coq, as of version 8.2, does not seem to properly support mutually
   recursive co-inductive and inductive definitions; that is, it allows these
   definitions to be made (up to a parameterization trick), but it does not
   allow useful proofs to be made about them, because the guardedness check
   for cofix is too restrictive and does not allow a call within a fixed
   point. *)

(* As a result, we abandon the idea of a co-inductive definition, and use a
   purely inductive definition, based on an approximation up to depth [n].
   In this, we follow Nakano's paper proof. *)

(* The predicate [hnf_where P t] means that the term [t] -- which is assumed
   to be a head normal form -- has components that satisfy [P]. *)

Inductive hnf_where (P : trm -> Prop) : trm -> Prop :=
| hnf_where_var:
    forall x,
    hnf_where P (trm_fvar x)
| hnf_where_abs:
    forall L t,
    (forall x, x \notin L -> hnf_where P t) ->
    hnf_where P (trm_abs t)
| hnf_where_app:
    forall t1 t2,
    hnf_where P t1 ->
    P t2 ->
    hnf_where P (trm_app t1 t2).

